#property version   "1.00"
#include "miniNeuralNetMV.mqh";
input string note0="----";//training 
input int batch_size=50000;//samples per training batch 
input bool reset_stats_after_batch=true;//reset stats after batch
input double base_learning_rate=0.001;//base learning rate 
input string note1="----";//display stoff:
input int px=10;//position x
input int py=100;//position y
input int displaySize=300;//display size
input color bgColor=C'200,30,30';//background color
input color isColor=clrWhite;//"is" color (or 1 or "on")
input color isNotColor=clrBlack;//"is not" color (or 0 or "off")
snn_mv_net net;
string system_tag="miniNN_";
uint pixels[];
union ucolor{
uint col;
uchar argb[4];
};
ucolor _is,_isnot;
bool NetBusy=false;
int OnInit()
  {
  ObjectsDeleteAll(ChartID(),system_tag);
  net.reset();
  net.setup(2);
  //setup the network for xor
    //layer 1 connects to input layer , 2 nodes 
    net.add_layer(2,2);
    //layer 2 connects to layer 1 , 2 nodes
    net.add_layer(2,2);
    //output layer connects to layer 2 , 1 node
    net.add_layer(1,2);
  //lets setup the display 
    string name=system_tag+"_display";
    ResourceFree("::DISPLAY");
    ObjectCreate(ChartID(),name,OBJ_BITMAP_LABEL,0,0,0);
    ObjectSetInteger(ChartID(),name,OBJPROP_XSIZE,displaySize);
    ObjectSetInteger(ChartID(),name,OBJPROP_YSIZE,displaySize);
    ObjectSetInteger(ChartID(),name,OBJPROP_YDISTANCE,py);
    ObjectSetInteger(ChartID(),name,OBJPROP_XDISTANCE,px);
    ArrayResize(pixels,displaySize*displaySize,0);
    ArrayFill(pixels,0,ArraySize(pixels),ColorToARGB(bgColor,255));
    ResourceCreate("DISPLAY",pixels,displaySize,displaySize,0,0,displaySize,COLOR_FORMAT_ARGB_NORMALIZE);
    ObjectSetString(ChartID(),name,OBJPROP_BMPFILE,"::DISPLAY");
    ChartRedraw();
    //cool , now we need to get the rgb values for on and off colors because we'd like to grade the result visually 
      _is.col=ColorToARGB(isColor,255);
      _isnot.col=ColorToARGB(isNotColor,255);
      NetBusy=false;
    //and create a timer 
      EventSetMillisecondTimer(44);
  return(INIT_SUCCEEDED);
  }

void OnTimer(){
if(!NetBusy)
  {
  NetBusy=true;
  
  //train a batch
    double features[],results[];
    ArrayResize(features,2,0);
    ArrayResize(results,1,0);
    /*
    what are these ? these are the features (inputs) of a sample and the correct answer ! (results)
    the results has the size of 1 because we have 1 output node
    */
    for(int i=0;i<batch_size;i++)
     {
     //first create xor inputs 0 or 1 and the expected result 
       int r=MathRand();
       if(r>16000){features[0]=1.0;}else{features[0]=0.0;}
           r=MathRand();
       if(r>16000){features[1]=1.0;}else{features[1]=0.0;}
     //okay lets fix the result too (xor)  
       if(features[0]!=features[1]){results[0]=1.0;}else{results[0]=0.0;}
     //lets pass the features through the network
       net.feed_forward(features);
     //and lets calculate the loss 
       net.calculate_loss(results);   
     //and lets backpropagate the error (we are not adjusting -learning- anything yet)
       net.back_propagation_same_features();  
     } 
     //"learn"
     double max_loss=net.get_max_loss_per_sample();
     double cur_loss=net.get_current_loss_per_sample();
     double a=(cur_loss/max_loss)*base_learning_rate;
     net.adjust(a);
     if(reset_stats_after_batch){net.reset_loss();} 
     string comm="Loss ("+cur_loss+")/("+max_loss+")";
     //tests
     features[0]=0.0;features[1]=0.0;
     net.feed_forward(features);
     comm+="\n[0][0]="+net.get_output(0);
     features[0]=1.0;features[1]=0.0;
     net.feed_forward(features);
     comm+="\n[1][0]="+net.get_output(0);     
     features[0]=0.0;features[1]=1.0;
     net.feed_forward(features);
     comm+="\n[0][1]="+net.get_output(0);     
     features[0]=1.0;features[1]=1.0;
     net.feed_forward(features);
     comm+="\n[1][1]="+net.get_output(0);     
     features[0]=0.5;features[1]=0.5;
     comm+="\n[0.5][0.5]="+net.get_output(0);
     Comment(comm);
     
     



     //update display 
       //we will treat the x axis as input[0] from 0 to 1.0
       //we will treat the y axis as input[1] from 0 to 1.0 
         ucolor painter;
         painter.argb[3]=255;
         for(int y=0;y<displaySize;y++){
         for(int x=0;x<displaySize;x++){
            //the hypothetical input [0] would be the x by the display size-1 
              features[0]=((double)x)/((double)(displaySize-1));
              features[1]=((double)y)/((double)(displaySize-1));
            //pass through and receive a value 0.0 to 1.0
              net.feed_forward(features);
              double output=net.get_output(0);
              /*
              now 0 is off 1 is on , duh , 
              % of off color is 1-output
              % of on  color is output
              */
              //channels
                
                for(int c=0;c<=2;c++){
                painter.argb[c]=(uchar)(MathFloor(((double)_isnot.argb[c])*(1.0-output))+MathFloor(((double)_is.argb[c])*(output)));
                } 
                             
              //set color 
                pixels[y*displaySize+x]=painter.col;
            }
         }
         //update resource
         ResourceCreate("DISPLAY",pixels,displaySize,displaySize,0,0,displaySize,COLOR_FORMAT_ARGB_NORMALIZE);
         ChartRedraw();
               
  NetBusy=false;
  }
}

void OnDeinit(const int reason)
  {
  ObjectsDeleteAll(ChartID(),system_tag);
  net.reset();
  }

void OnTick()
  {
   
  }

